import { RECEIVE_PROPS } from 'melody-component';

var _typeof = typeof Symbol === "function" && typeof Symbol.iterator === "symbol" ? function (obj) {
  return typeof obj;
} : function (obj) {
  return obj && typeof Symbol === "function" && obj.constructor === Symbol && obj !== Symbol.prototype ? "symbol" : typeof obj;
};

/**
 * Copyright 2017 trivago N.V.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS-IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
/**
 * Wraps a series of redux compatible middlewares into a Melody Component Mixin.
 *
 * @param {...Function} middlewares
 */
var applyMiddleware = function applyMiddleware() {
    for (var _len = arguments.length, middlewares = Array(_len), _key = 0; _key < _len; _key++) {
        middlewares[_key] = arguments[_key];
    }

    return function (proto) {
        var _componentDidInitialize = proto.componentDidInitialize;
        return {
            componentDidInitialize: function componentDidInitialize() {
                _componentDidInitialize.call(this);
                // and dispatch
                var next = this.dispatch;
                var i = middlewares.length;
                this.dispatch = function (action) {
                    return next(action);
                };
                while (i--) {
                    var curr = middlewares[i];
                    if (curr) {
                        next = curr(this)(next);
                    }
                }
            }
        };
    };
};

/**
 * Wraps an object into a reducer function so that each property of the object should
 * correspond to the action type that it handles.
 *
 * @param {Object} pattern The object that should be used to delegate to the correct reducer.
 * @param {Object?} initialState An optional state that should be used on initialization.
 */
var createActionReducer = function createActionReducer(pattern, initialState) {
    return function (state, action) {
        if (action.type && pattern[action.type]) {
            return (0, pattern[action.type])(state || initialState, action) || state || initialState;
        }
        return state || initialState;
    };
};

var mapDispatchToStateWithProps = function mapDispatchToStateWithProps(fn) {
    return function (state, action) {
        if (action.type === 'MELODY/RECEIVE_PROPS') {
            return Object.assign({}, state, fn(action.meta.dispatch, action.payload));
        }
        return state;
    };
};

var DISPATCH_PROPS = 'MELODY/DISPATCH_TO_STATE';

var mapDispatchToState = function mapDispatchToState(fn) {
    return function (state, action) {
        if (action.type === 'MELODY/RECEIVE_PROPS') {
            if (state && !state[DISPATCH_PROPS]) {
                var dispatchMap = fn(action.meta.dispatch);
                return Object.assign({}, state, dispatchMap, {
                    'MELODY/DISPATCH_TO_STATE': dispatchMap
                });
            }
        }
        return state;
    };
};

/**
 * Returns a function that will dispatch the return value of the given action creator
 * to the given dispatch function.
 *
 * Can be used to create an action that is always dispatched to the same store or component.
 *
 * @param {Function} action
 * @param {Function} dispatch
 */
var bindActionToDispatch = function bindActionToDispatch(action, dispatch) {
    return function () {
        return dispatch(action.apply(undefined, arguments));
    };
};

var wrapObjectToDispatch = function wrapObjectToDispatch(dispatchToState) {
    return function (dispatch) {
        var keys = Object.keys(dispatchToState);
        var mappedDispatchers = {};
        var i = 0;
        for (var len = keys.length; i < len; i++) {
            var actionCreator = dispatchToState[keys[i]];
            if (typeof actionCreator === 'function') {
                mappedDispatchers[keys[i]] = bindActionToDispatch(actionCreator, dispatch);
            }
        }
        return mappedDispatchers;
    };
};

/**
 * Returns a function that can be used to inject dispatchers into the state
 * of a component.
 *
 * Usually used together with [reduce-reducers](https://github.com/acdlite/reduce-reducers).
 *
 * @param {Function|Object} dispatchToState The dispatch reducer
 */
function dispatchToState(dispatchToState) {
    if ((typeof dispatchToState === 'undefined' ? 'undefined' : _typeof(dispatchToState)) === 'object') {
        return mapDispatchToState(wrapObjectToDispatch(dispatchToState));
    }
    var dependsOnProps = dispatchToState.length === 2;
    if (dependsOnProps) {
        return mapDispatchToStateWithProps(dispatchToState);
    } else {
        return mapDispatchToState(dispatchToState);
    }
}

var defaultReducer = function defaultReducer() {
    var state = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : {};
    var _ref = arguments[1];
    var type = _ref.type,
        payload = _ref.payload;

    if (type === RECEIVE_PROPS) {
        return Object.assign({}, state, payload);
    }
    return state;
};

/**
 * Returns a reducer that exposes certain functions of a component to the state
 *
 * @param {Array} list List of property names that should be exposed to the state
 * @param {Function} reducer Component reducer that will be wrapped
 */
function exposeToState() {
    var list = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : [];
    var reducer = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : defaultReducer;

    return function (state, action) {
        var type = action.type,
            meta = action.meta;

        var result = reducer(state, action);

        // Check if functions have already been merged to state;
        var hasExpose = true;
        for (var i = 0, l = list.length; i < l; i++) {
            var key = list[i];
            if (!result[key]) {
                hasExpose = false;
                break;
            }
        }

        if (hasExpose || type !== RECEIVE_PROPS) {
            return result;
        }

        var expose = list.reduce(function (acc, key) {
            var prop = meta[key];
            if (!prop) {
                throw new Error('Property `' + key + '` was not found on the component.');
            }
            if (typeof prop !== 'function') {
                throw new Error('Property `' + key + '` is not a function. Only functions can be exposed to the state.');
            }
            acc[key] = prop.bind(meta);
            return acc;
        }, {});

        return Object.assign({}, result, expose);
    };
}

export { applyMiddleware, createActionReducer, bindActionToDispatch, dispatchToState, exposeToState };
